from typing import Tuple
import pytest
import numpy as np
import xarray as xr
from mempy.model import (
    RED_IT, RED_SD, RED_IT_DA, RED_SD_DA, RED_IT_IA, RED_SD_IA, Reduced,
    BufferGUTS_SD, BufferGUTS_IT, BufferGUTS_SD_DA, BufferGUTS_SD_CA, BufferGUTS_IT_DA, 
    BufferGUTS_IT_CA
)
from mempy.input_data import read_exposure_survival
from guts_base import PymobSimulator

# results are from Bürger and Focks 2025 (https://doi.org/10.1093/etojnl/vgae058) 
# supplementary material (Tab. 5.3)
OPENGUTS_ESTIMATES = dict(
    red_sd = xr.Dataset(dict(kd=0.712, m=2.89, b=0.619, hb=0.008)).to_array().sortby("variable"),
    red_it = xr.Dataset(dict(kd=0.789, m=5.36, beta=5.08, hb=0.025)).to_array().sortby("variable"),
    red_sd_da = None,
    red_sd_ia = None,
    red_it_da = None,
    red_it_ia = None,
    bufferguts_it = None,
    bufferguts_sd = None,
    bufferguts_sd_ca = None,
    bufferguts_sd_da = None,
    bufferguts_it_ca = None,
    bufferguts_it_da = None,
    # TODO: Define true parameters for red_sd_da
)

def read_data(file):
    data = read_exposure_survival(
        "data/", file, 
        survival_name="Survival",
        exposure_name="Exposure",
        visualize=False,
        with_raw_exposure_data=True
    )

    exposure_funcs, survival_data, num_expos, exposure_data = data
    info_dict = {}

    return exposure_funcs, survival_data, num_expos, info_dict, exposure_data


def construct_sim(dataset: Tuple, model: type, output_path="results/testing"):
    """Helper function to construct simulations for debugging"""
    _, survival_data, num_expos, _, exposure_data = read_data(file=dataset)

    if model in (RED_IT, RED_SD, BufferGUTS_SD, BufferGUTS_IT):
        _model = model()
    else:
        _model = model(num_expos=num_expos)

    sim = PymobSimulator.from_mempy(
        exposure_data=exposure_data,
        survival_data=survival_data,
        model=_model,
        output_directory=output_path
    )

    return sim


@pytest.fixture(params=[
    ("ringtest_A_IT.xlsx", RED_IT),
    ("ringtest_A_SD.xlsx", RED_SD),
    ("Fit_Data_Cloeon_final.xlsx", RED_SD_DA,),
    ("Fit_Data_Cloeon_final.xlsx", RED_SD_IA,),
    ("Fit_Data_Cloeon_final.xlsx", RED_IT_DA,),
    ("Fit_Data_Cloeon_final.xlsx", RED_IT_IA,),
    ("osmia_contact_synthetic.xlsx", BufferGUTS_SD,),
    ("osmia_contact_synthetic.xlsx", BufferGUTS_IT,),
    ("osmia_multiexpo_synthetic.xlsx", BufferGUTS_SD_CA,),
    ("osmia_multiexpo_synthetic.xlsx", BufferGUTS_SD_DA,),
    ("osmia_multiexpo_synthetic.xlsx", BufferGUTS_IT_CA,),
    ("osmia_multiexpo_synthetic.xlsx", BufferGUTS_IT_DA,),
])
def dataset_and_model(request) -> Reduced:
    yield request.param


# Derive simulations for testing from fixtures
@pytest.fixture
def sim(dataset_and_model, tmp_path):
    dataset, model = dataset_and_model
    yield construct_sim(dataset=dataset, model=model, output_path=tmp_path)


# run tests with the Simulation fixtures
def test_setup(sim):
    """Tests the construction method"""
    assert True


def test_simulation(sim):
    """Tests if a forward simulation pass can be computed"""
    # sim.dispatch_constructor()
    evaluator = sim.dispatch()
    evaluator()
    evaluator.results

    assert True

@pytest.mark.slow
@pytest.mark.parametrize("backend", ["numpyro"])
def test_inference(sim: PymobSimulator, backend):
    """Tests if prior predictions can be computed for arbitrary backends"""

    sim.set_inferer(backend)

    sim.prior_predictive_checks()
    sim.inferer.run()

    sim.posterior_predictive_checks()

    sim.config.report.debug_report = True
    sim.report()

    # test if inferer converged on the true estmiates
    pymob_estimates = sim.inferer.idata.posterior.mean(("chain", "draw")).to_array().sortby("variable")
    openguts_estimates = OPENGUTS_ESTIMATES[sim.config.simulation.model.lower()]

    if openguts_estimates is None:
        # this explicitly skips testing the results, since they are not available,
        # but does not fail the test.
        pytest.skip()

    np.testing.assert_allclose(pymob_estimates, openguts_estimates, rtol=0.05, atol=0.1)


if __name__ == "__main__":
    # test_simulation(sim=construct_sim(data_ringtest, model=RED_IT))
    test_inference(sim=construct_sim(dataset="Fit_Data_Cloeon_final.xlsx", model=BufferGUTS_SD_DA), backend="numpyro")
    # test_inference(sim=construct_sim(dataset="Fit_Data_Cloeon_final.xlsx", model=RED_IT_IA), backend="numpyro")
    # test_inference(sim=construct_sim(dataset="paula_data/Dimethoate_Clothianidin_Propiconazole_constant_plsed - muegL.xlsx", model=RED_IT_IA), backend="numpyro")


'''
This file contains the actual model equations and function to solve them and calculate the likelihood of given parametersets.
Models are strucuted in the following way:

                        
Top level:              shared parent class
                            | (inherits)
Intermediate level:     model-group specific derived class
                            | (inherits)
Bottom level:           model specific derived class


Important functions
-------------------
__init__ (Top level and Bottom level)
    Defines the variables of the specific model equations in a dict, containg starting values for fitting, ranges, names etc.
lp_X() (Top level)
    Calculates the LP_X for a given treatment and parameter set
_rhs (Bottom level)
    "right hand side" of the model specific ODEs to be integrated.
_solve (Top level and Bottom level)
    Solver for the model ODEs. For most SD model variants this is implemented in the Top level class. In case of particularities like the 
    dependance on the maximum damage up to this point in time in IT model variants this must be overwritten in the Bottom level class. This
    is because the scipy.integrate functions cannot handle anything but clean model ODEs in the _rhs()-methods.
plot (Intermediate level)
    Plots the model solutions with the fitted parameter values and, if given, the input-survival-data.
nrmse (Intermediate level)
    Calculates the Normalized Root Mean Square Error
negloglike (Intermediate level)
    Calculation of the negative logarithmic likelihood for parameter fitting.
mortality_loglogistic (Top level)
    Calculate mortality from cumulative log-logistic distribution

'''
import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate as scipyInt
import pandas as pd
from scipy import interpolate

try:
    import jax
    import jax.numpy as jnp
    jax.config.update("jax_enable_x64", True)
except ModuleNotFoundError:
    pass

import mempy.input_data as input_data
from mempy.survival_models import (
    conditional_survival_hazard_error_model,
    conditional_survival_error_model
)

'''
----------------------------------------------------------------------------------------------------------------------------------------------------
TOP LEVEL: 
One shared class for all guts model variants defining necessary function and initialising important parameters.
----------------------------------------------------------------------------------------------------------------------------------------------------
''' 
class Model:
    extra_dim = None

    def __init__(self):
        '''
        Initialises some important variables with standard values. In case of a more unusual guts variant these must be overwritten in
        the  __init__-method of the variant specific derived class.
        '''
        self.params_info = {}
        self._params_info_defaults = {
            "initial": 1.0, 
            "name": None,
            "min": None,
            "max": None,
            "prior": None,
            "dims": None,
            "vary": True,
        }
        self.comp_info = [
            'Scaled Damage',
            'Survival'
        ]
        self._numCLike = 1
        self._numSLike = 1
        self.num_expos = 1

        self.state_variables = {
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = None

    def mortality_loglogistic(self, x, alpha, beta):
        '''Calculate mortality from cumulative log-logistic distribution for IT model variants

        Parameters
        ----------
        x
            the scaled damage
        alpha
            the median threshold value
        beta
            the shape parameter

        Returns
        ----------
            cumulative log-logistic mortality given the parameters
        '''
        if x > 0:
            return 1.0 / (1 + (x/alpha)**-beta)
        else:
            return 0.0
        
    def LP_X(self, params, exposure_funcs, treatment, accuracy = 0.001, X = 50, visualize = True):
        '''
        Calculates the LP_X for a given treatment and parameter set

        Parameters
        ----------
        params: dict
            dict containing parameter values suitable for this model
        exposure_funcs: dict
            dict containing the exposure (see mempy.input_data)
        treatment: string
            name of the treatment to be used/ key of this treatment in the exposure_funcs dict
        accuracy: float
            accuracy of the LP_X
        X: float
            the percentage of survival to be reached in the last timestep
        visualize: boolean
            wheter or not the result should be plotted

        '''
        left = 0
        right = 1000000
        mid = None
        lp = X/100
        exp_f = {}
        for sub in exposure_funcs:
            exp_f[sub] = {}
            exp_f[sub][treatment] = {}
            knots = exposure_funcs[sub][treatment].get_knots()
            coeffs = exposure_funcs[sub][treatment].get_coeffs()
            exp_f[sub][treatment]['knots'] = knots
            exp_f[sub][treatment]['coeffs'] = coeffs
        
        def multiply_exp(exp_or, treat, factor):
            exp_new = {}
            for sub in exp_or:
                exp_new[sub] = {}
                original_coeffs = exp_or[sub][treat]['coeffs']
                knots = exp_or[sub][treat]['knots']
                end = knots[-1]
                exp_new[sub][treat] = exposure_funcs[sub][treat]
                exp_new[sub][treat + '_lp'] = interpolate.InterpolatedUnivariateSpline(knots, original_coeffs*factor, k=1)
            return end, exp_new
        
        while left <= right:
            mid = left + (right - left) / 2

            y0 = np.zeros(self._numCLike + self._numSLike, dtype=np.double)
            for j in range(self._numSLike):
                y0[-(j+1)]=1
        
            end, exp = multiply_exp(exp_f, treatment, mid)
            res = self._solve(params,y0,[0,end],input_data.get_xc(treatment + '_lp', exp))[-1][-1]
            if res + accuracy > lp and res - accuracy < lp:
                if visualize:
                    print(f'The LP_{X} is {mid}, leading to a final survival of {res}')
                    self.plot(params, exp, share_axis = True, suptitle = f'Model results with LP_{X}.')
                return mid
            if res > lp:
                left = mid
            else:
                right = mid
    
    def _rhs(self):
        '''
        Contains the model ODEs for a specific guts model variant. Must be overwritten in the model variant specific class.
        '''
        raise NotImplementedError('Must be implement in derived class (Bottom level)')
    
    def _rhs_jax(self):
        '''
        Contains the model ODEs for a specific guts model variant in jax syntax. Must be overwritten in the model variant specific class.
        '''
        raise NotImplementedError('Must be implement in derived class (Bottom level)')
    
    def negloglike(self):
        '''
        Calculates the negative logarithmic likelihood of a parameterset with respect to given training data. Must be implemented
        in a intermediate level derived class to specify the calculation depending on the specific model structure. For the standard 
        guts-reduced models this is the model.Reduced-class.
        '''
        raise NotImplementedError('negloglike()-function: must be implement in derived class (Intermediate level)')
    

    def nrmse(self):
        '''
        Calculates the negative logarithmic likelihood of a parameterset with respect to given training data. Must be implemented
        in a intermediate level derived class to specify the calculation depending on the specific model structure. For the standard 
        guts-reduced models this is the model.Reduced-class.
        '''
        raise NotImplementedError('nrmse().function: must be implement in derived class (Intermediate level)')
    

    def plot(self, params, exposure_funcs, observed_data = pd.DataFrame(), share_axis = True, suptitle = None):
        '''
        Plots the model solutions with the fitted parameter values and, if given, the input-survival-data.
        '''
        raise NotImplementedError('plot()-function: must be implement in derived class (Intermediate level)')


    def _solve(self, params, y0, times, xc):
        '''
        Solver for model ODEs. For most SD model variants this is implemented in the Top level class. In case of particularities like the 
        dependance on the maximum damage up to this point in time in IT model variants this must be overwritten in the Bottom level class. This
        is because the scipy.integrate functions cannot handle anything but clean model ODEs in the _rhs()-methods.

        Parameters
        ----------
        params
            A set of parameter values suitable for this model instance
        y0
            The initial conditions for the compartments
        times
            The timepoints at which the model should be evaluated
        xc
            A set of functions containing the time-dependant exposure profile
        
        Returns
        ----------
        result
            An array of lists containing the values for all model compartments at a specific timepoint.
        '''
        
        int_ode = scipyInt.ode(self._rhs).set_integrator('lsoda')
        int_ode.set_initial_value(y0)
        int_ode.set_f_params(params, xc)

        result = [y0]
        for t in times[1:]:
            int_ode.integrate(t)
            result.append(int_ode.y)

        return np.array(result)

    @staticmethod
    def _solver_post_processing(results, time, interpolation):
        return results

'''
----------------------------------------------------------------------------------------------------------------------------------------------------
INTERMEDIATE LEVEL: 
Different classes that implement likelihood calculation for different groups of guts models. 
All clases inherit from the model.Model-class and are used as parent for their model specific derived classes.
----------------------------------------------------------------------------------------------------------------------------------------------------
''' 
class Reduced(Model):
    extra_dim = "substance"
    '''
    This class contains the calculation of the negative logarithmic likelihood for all standard reduced guts models. The nll is
    calculated solely considering the survival.
    For different models, like e.g. the full guts model a class parallel to this is needed, implementing the likelihood calculation. This class
    is then used as parent for the specific model class.
    '''

    def negloglike(self, params, datasets):
        '''
        Calculates the negative logarithmic likelihood of a certain set of parameters and a number of treatments.

        This implements equation (3.41) from the guts paper (p. 57 f)

        Parameters
        ----------
        params: dict
            A suitable set of parameterset for the this model instance.
        datasets: dict
            The training data in the form of datasets (see input_data.create_datasets()).

        Returns
        ----------
        negloglikelihood: float
            The negative logarithmic likelihood of the given parameterset (params) with respect to the given training data.

        '''

        y0 = np.zeros(self._numCLike + self._numSLike)
        for i in range(self._numSLike):
            y0[-(i+1)] = 1.0

        negloglikelihood = 0
        for dataset in datasets:
            xc = tuple(dataset[:-1])

            expected_survival = dataset[-1]
            times = expected_survival.index.values
            expected_survival = np.array(expected_survival)

            y = self._solve(params, y0, times, xc)
            predicted_survival = np.array(y[:,-1], dtype = np.double)

            # Tail regularization parameter to avoid errors due to numercial precision limits
            eps = 1e-9
            predicted_survival[predicted_survival<eps] = eps

            data_diff = np.diff(expected_survival) * -1
            model_diff = np.diff(predicted_survival) * -1

            model_diff[model_diff<eps] = eps

            temp_ll = np.sum(data_diff * np.log(model_diff)) + expected_survival[-1] * np.log(predicted_survival[-1])

            if np.isnan(temp_ll):
                raise ValueError('Likelihood cannot be NaN')
            
            negloglikelihood -= temp_ll

        return negloglikelihood


    def nrmse(self, params, exposure_funcs, survival_data, with_akaike = False):
        '''
        Calculates the Normalized root mean square error of the model

        Parameters
        ----------
        params: dict
            the model parameter values to be used
        exposure_funcs: dict
            dictionary of function representing external concentrations
        survival_data: pandas.DataFrame
            the observed survival data to compare the model results to
        with_akaike: boolean
            wheter or not the Akaike information criterium is to be calucalted as well

        Returns
        ----------
        nrmse: float
            the nrmse of the model
        aic: float
            the aic of the model
        '''
        
        RMSE_top = 0
        RMSE_bot = 0
        all_observed = []
        treatments = survival_data.columns
        for i, treatment in enumerate(treatments):
            survival = survival_data[treatment].dropna()
            times = survival.index  
            xc = input_data.get_xc(treatment, exposure_funcs)
            
            #adding a 0 per damage compartment and the number of organisms per survival compartment to the initial values array
            s = survival.iloc[0]
            y0 = np.zeros(self._numCLike + self._numSLike, dtype=np.double)
            for i in range(self._numSLike):
                y0[-(i+1)]=s

            # Evaluate model at data time points with present parameter set
            y = self._solve(params, y0, times, xc)
            observed = survival.to_numpy()
            modelled = y[:,-1]
            RMSE_top += np.sum((modelled - observed)**2) 
            RMSE_bot += len(observed) 
            all_observed.extend(observed)
        RMSE = np.sqrt(RMSE_top/RMSE_bot) 
        NRMSE = RMSE/np.mean(np.array(all_observed)) 

        if not with_akaike:
            return NRMSE
        else:

            datasets = input_data.create_datasets(exposure_funcs, survival_data)
            neglog = self.negloglike(params, datasets)
            k = len(params)
            n = len(survival_data.columns)
            # different definitions of the AIC 
            aic = 2 * neglog + 2*k #+ (2*k**2+2*k)/(n-k-1)
            #aic = RMSE * np.log(RMSE_top / RMSE_bot) + 2*len(self.params)
            #aic =  n * np.log(RMSE_top/n) + 2*k + (2*k**2+2*k)/(n-k-1)
            return NRMSE, aic
        
    
    def plot(self, params, exposure_funcs, observed_data = pd.DataFrame(), share_axis = True, suptitle = None, return_data = False):
        '''
        Plots the model solutions with the fitted parameter values and, if given, the input-survival-data.

        Parameters
        ----------
        params: dict
            The parameter values appropriate for the used model
        exposure_funcs: dict
            Dict of exposures and treatments containing functions representing the external concentrations.
        observed_data: pandas.DataFrame
            Pandas.DataFrame containing the observed survival for the treatments and timepoints.
        share_axis: boolean
            Wheter or not the plots for each compartment should share their y-axis to achieve better visual comparability.
        suptitle: string
            The Title for the figure.
        return_data: boolean
            Wheter or not the simulated data should be returned for further use

        Returns
        ----------
        res_df: pd.DataFrame
            The simulated data for the given exposures and parameters for further use
        '''
        naming_expf = {}
        for expf in exposure_funcs:
            naming_expf = exposure_funcs[expf]
            break
        timing_expf = None
        for treat in naming_expf:
            timing_expf = naming_expf[treat]
            break
        starttime = timing_expf.get_knots()[0]
        endtime = timing_expf.get_knots()[-1]

        times = np.linspace(starttime, endtime, int((endtime-starttime)*100))
        # add the timepoints from the observed data into the timepoints to be simulated to ensure comparability
        if not observed_data.empty:
            times = np.insert(times, -1, np.array(observed_data.index))
            times = np.unique(times)

        if suptitle == None:
            headline = 'Model output'
        else:
            headline = suptitle
        numCLike = self._numCLike
        numSLike = self._numSLike

        num_expos = self.num_expos
        columns = 0
        for sub in exposure_funcs:
            columns = len(exposure_funcs[sub])
        rows = numCLike + numSLike + num_expos
        fig, axis = plt.subplots(rows, columns, figsize=(columns*5, rows*5))
        fig.tight_layout()

        res = {}
        for i, treat in enumerate(naming_expf.keys()):
            #adding a 0 per damage compartment and the number of organisms per survival compartment to the initial values array
            if not observed_data.empty:
                s = observed_data[treat].iloc[0]
            else:
                s = 1
            y0 = np.zeros(numCLike + numSLike, dtype=np.double)
            for j in range(numSLike):
                y0[-(j+1)]=s
            
            xc = input_data.get_xc(treat, exposure_funcs)
            y = self._solve(params, y0, times, xc)
            res[treat] = y

        if share_axis:
            max_y_ax = []
            for sub in exposure_funcs:
                max_exp_temp = 0
                for treat in exposure_funcs[sub]:
                    temp_max = max(exposure_funcs[sub][treat](times))
                    if max_exp_temp < temp_max:
                        max_exp_temp = temp_max
                max_y_ax.append(max_exp_temp)
            
            for n in np.arange(0,numCLike+numSLike):
                max_comp_n = 0
                for treat in res:
                    y = res[treat]
                    temp_max = max(y.transpose()[n])
                    if max_comp_n < temp_max:
                        max_comp_n = temp_max
                max_y_ax.append(max_comp_n)

        for i, treat in enumerate(naming_expf.keys()):
            y = res[treat]
            for k, sub in enumerate(exposure_funcs):
                ax = axis[k][i]
                ax.plot(times, exposure_funcs[sub][treat](times))
                if share_axis:
                    ax.set(title = f'Exposure {sub} {treat}', ylim = (-max_y_ax[k]/50, max_y_ax[k]+max_y_ax[k]/50))
                else:
                    ax.set(title = f'Exposure {sub} {treat}')
            for l, k in enumerate(np.arange(num_expos,num_expos+numCLike,1)):
                ax = axis[k][i]
                ax.plot(times, y.transpose()[l])
                if share_axis:
                    ax.set(title = f'{self.comp_info[l]} {treat}', ylim = (-max_y_ax[k]/50, max_y_ax[k]+max_y_ax[k]/50))
                else:
                    ax.set(title = f'{self.comp_info[l]} {treat}')
            for m, k in enumerate(np.arange(num_expos+numCLike, num_expos+numCLike+numSLike, 1)):
                ax = axis[k][i]
                ax.plot(times, y.transpose()[m+numCLike])
                if not observed_data.empty:
                    ax.scatter(observed_data[treat].index, observed_data[treat])
                if share_axis:
                    ax.set(title = f'{self.comp_info[m+numCLike]} {treat}', ylim = (-max_y_ax[k]/50, max_y_ax[k]+max_y_ax[k]/50))
                else:
                    ax.set(title = f'{self.comp_info[m+numCLike]} {treat}')
        fig.suptitle(headline, y=1.05, fontsize = 20)
        plt.show()

        if return_data:
            res_df = pd.DataFrame()
            res_df.insert(0, 'Time', times)
            res_df.set_index('Time', drop = True, inplace= True)

            for treat in res:
                assert len(res[treat].transpose()) == len(self.comp_info)
                for i, comp in enumerate(self.comp_info):
                    res_df.insert(len(res_df.columns), f'{comp}_{treat}', res[treat].transpose()[i])

            return res_df



'''
----------------------------------------------------------------------------------------------------------------------------------------------------
BOTTOM LEVEL: 
Classes that implement specific model ODEs (and if necessary solvers for them). They inherit from a suitable parent class from the
intermediate level to ensure proper likelihood calculation.
----------------------------------------------------------------------------------------------------------------------------------------------------
'''
class RED_SD(Reduced):
    '''
    the GUTS-RED-SD (Stoachastic Death) model
    '''
    def __init__(self):
        super().__init__()

        self.params_info['kd'] = {'name':'kd', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1.0,s=3)"}
        self.params_info['b'] = {'name':'b', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1.0,s=3)"}
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1.0,s=3)"}
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-8, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1.0,s=3)"}

        self.state_variables = {
            "D": {"dimensions": ["id", "time"], "observed": False, "y0": [0.0] * self._numCLike},
            "H": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model


    def _rhs(self, t, y, params, xc):
        '''
        Right-hand side DGLs 
        xc: list of functions (interpolation over time)
        y: actual model state
        '''
        xc = xc[0]
        dy = y.copy()

        # Damage
        dy[0] = params['kd'] * (xc(t) - y[0])

        # Hazard rate
        h_substance = params['b'] * max(0, y[0] - params['m'])
        h = h_substance + params['hb']

        # Survival
        dy[1] = -h * y[1]

        return dy

    @staticmethod
    def _rhs_jax(t, y, x_in, kd, b, m, hb):
        D, H = y
        dD_dt = kd * (x_in.evaluate(t) - D)

        # Hazard rate
        dH_dt = b * jnp.maximum(0.0, D - m) + hb

        return dD_dt, dH_dt

    @staticmethod
    def _solver_post_processing(results, time, interpolation):
        results["survival"] = jnp.exp(-results["H"])
        results["exposure"] = jax.vmap(interpolation.evaluate)(time)
        return results


class RED_IT(Reduced):
    '''
    the GUTS-RED-IT (Individual Tolerance) model
    '''

    def __init__(self):
        super().__init__()
        self.params_info['kd'] = {'name':'kd', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1,s=3)"}
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1,s=3)"}
        self.params_info['beta'] = {'name':'beta', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1,s=3)"}
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-8, 'max':1.0e3, 'initial':1.0, 'vary':True, "prior": "lognorm(scale=1,s=3)"}
        
        self.state_variables = {
            "D": {"dimensions": ["id", "time"], "observed": False, "y0": [0.0] * self._numCLike},
            "H": {"dimensions": ["id", "time"], "observed": False},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model

    def _rhs(self, t, y, params, xc):
        xc = xc[0]
        dy = y.copy()

        # Damage
        dy[0] = params['kd']*(xc(t) - y[0])

        dy[1] = -99

        return dy

    def _solve(self, params, y0, times, xc):        
        #Solve uptake equation
        y = super()._solve(params, y0,  times, xc)

        #Calculate individual tolerance survival
        for i, ystep in enumerate(y):
            if i == 0:
                continue

            t = times[i]

            #Maximum damage to this point
            D_wm = np.max(y[:i+1,0])

            #Calculate survival from ratio of max damage to tolerance threshold.
            F = self.mortality_loglogistic(D_wm, params['m'], params['beta'])
            surv = y[0, -1] * (1.0 - F) * np.exp(-params['hb']*t)
            y[i, -1] = surv

        return y

    @staticmethod
    def _rhs_jax(t, y, x_in, kd):
        D, = y
        C = x_in.evaluate(t)

        dD_dt = kd * (C - D)

        return (dD_dt, )
    
    @staticmethod
    def _solver_post_processing(results, t, interpolation, m, beta, hb, eps):
        """
        TODO: Try alternative formulation. This is computationally simpler and numerically
        more stable:
        log S = log 1.0 + log (1.0 - F) + log exp -hb * t = 0 + log (1.0 - F) - hb * t
        """

        d_max = jnp.squeeze(jnp.array([jnp.max(results["D"][:i+1])+eps for i in range(len(t))]))
        F = jnp.where(d_max > 0, 1.0 / (1.0 + (d_max / m) ** -beta), 0)
        S = 1.0 * (jnp.array([1.0], dtype=float) - F) * jnp.exp(-hb * t)
        results["H"] = - jnp.log(S)
        results["survival"] = S
        results["exposure"] = jax.vmap(interpolation.evaluate)(t)

        return results
    
class Mixture(Reduced):
    extra_dim = "substance"

class RED_SD_DA(Mixture):
    '''
    the GUTS-RED-SD-DA model. This is the SD option of the damage addition guts mixtures model.
    '''
    def __init__(self, num_expos = 1):
        super().__init__()  
        self.num_expos = num_expos
        self._numCLike = num_expos + 1
        self._numSLike = 1

        self._likelihood_func_jax = conditional_survival_hazard_error_model

        self.state_variables = {
            "D": {"dimensions": ["id", "time", "substance"], "observed": False, "y0": [0.0] * num_expos},
            "H": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }


        ###substance specific params
        for i in range(self.num_expos):
            #dominant rate constants
            self.params_info[f'kd{i+1}'] = {'name':f'kd{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance"}
            if i == 0:
                #weight factors
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'initial':1.0, 'vary':False, "dims": "substance"}
            else:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance"}
        ###shared params
        #killing rate
        self.params_info['b'] = {'name':'b', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #median of the threshold distribution 
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #background mortality rate
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'Scaled damage {i+1}')
        self.comp_info.append('total scaled damage')
        self.comp_info.append('Survival')
    
    def _rhs(self, t, y, params, xc):
        
        dy = y.copy()
        
        # TODO: Paula/Leo: is this the same as before?
        #total damage from the last timestep
        # d_ges_last = y[0]

        # for i in range(self.num_expos):
        #     if i == 0:
        #         continue
        #     else:
        #         d_ges_last += y[i] * params[f'w{i+1}']
        
        d_ges_last = np.sum([params[f'w{i+1}'] * y[i] for i in range(self.num_expos)])

        #Damages
        for i in range(self.num_expos):
            dy[i] = params[f'kd{i+1}'] * (xc[i](t) - y[i])
        
        #Hazard rate; dependant of total damage
        hc = params['b'] * max(0, d_ges_last - params['m'])
        #Adding background mortality
        h = hc + params['hb']
        #survival
        dy[-1] = -h * y[-1]

        return dy
    
    @staticmethod
    def _rhs_jax(t, y, x_in, kd, w, b, m, hb):
        D, H = y
        dD_dt = kd * (x_in.evaluate(t) - D)

        # Hazard rate
        dH_dt = b * jnp.maximum(0.0, jnp.sum(w * D) - m) + hb

        return dD_dt, dH_dt

    @staticmethod
    def _solver_post_processing(results, time, interpolation):
        results["survival"] = jnp.exp(-results["H"])
        results["exposure"] = jax.vmap(interpolation.evaluate)(time)
        return results

    #solver for model ODEs, using scipy.integrate.ode.integrate with 'lsoda' integrator
    
    def _solve(self, params, y0, times, xc):
        
        #defining integrator, initial values and parameters
        int_ode = scipyInt.ode(self._rhs).set_integrator('lsoda')
        int_ode.set_initial_value(y0)
        int_ode.set_f_params(params, xc)
        
        #creating result data structures
        result = [y0]
        #actual integration; insertion of the total damage into the result-list
        for t in times[1:]:
            int_ode.integrate(t)
            res = int_ode.y

            #calculation of total damage for result array
            d_ges = res[0]
            for i in range(self.num_expos):
                if i == 0:
                    continue
                else:
                    d_ges += res[i] * params[f'w{i+1}']
            res[-2] = d_ges
            result.append(res)
        #returns sols; contains [damage1,...,damageN, damagesGES, survivalGES] as columns
        return np.array(result)   
    

class RED_IT_DA(Mixture):
    '''
    the GUTS-RED-IT-DA model. This is the IT option of the damage addition guts mixtures model.
    '''
    
    def __init__(self, num_expos = 1):
        super().__init__()   
        self.num_expos = num_expos
        self._numCLike = num_expos+1
        self._numSLike = 1

        # For the IT model this is equivalent to the hazard_error_model, because
        # IT calculates the hazard as -log(S)
        self._likelihood_func_jax = conditional_survival_error_model

        self.state_variables = {
            "D": {"dimensions": ["id", "time", "substance"], "observed": False, "y0": [0.0] * num_expos},
            "H": {"dimensions": ["id", "time"], "observed": False,},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }


        ### substance specific parameters
        for i in range(self.num_expos):
            #dominant rate constants
            self.params_info[f'kd{i+1}'] = {'name':f'kd{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance"}
            if i == 0:
                #weight factors
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'initial':1.0, 'vary':False, "dims": "substance"}
            else:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance"}
        ### shared parameters
        #median of the threshold distribution 
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #shape parameter
        self.params_info['beta'] = {'name':'beta', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #background mortality
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'scaled damage {i+1}')
        self.comp_info.append('total scaled damage')
        self.comp_info.append('survival')
            
    def _rhs(self, t, y, params, xc):

        dy = y.copy()
        
        for i in range(self.num_expos):
            dy[i] = params[f'kd{i+1}'] * (xc[i](t) - y[i])
        
        return dy 
        
    
    def _solve(self, params, y0, times, xc):

        int_ode = scipyInt.ode(self._rhs).set_integrator('lsoda')
        int_ode.set_initial_value(y0)
        int_ode.set_f_params(params, xc)
        
        #creating result data structures
        result = [y0]
        #actual integration; insertion of the total damage (class attribute) into the result-list
        for t in times[1:]:
            int_ode.integrate(t)
            res = int_ode.y
            
            #calculation of total damage for result array
            d_ges = np.sum([params[f'w{i+1}'] * res[i] for i in range(self.num_expos)])

            res[-2] = d_ges
            result.append(res)
               
        y = np.array(result)

        #sols/y now contains the values for [damage1,...,damageN, damagesGES] and placeholders for 
        #the survival which are calculated in the following since the dependance of the 
        #maximum total damage cannot be realised within the integrator
        
        #determine the maximum total damage for every timestep and calculate the resulting survival
        for i, ystep in enumerate(y):
            if i == 0:
                continue
            
            t = times[i]

            #Maximum total damage to this point
            D_wm = np.max(y[:i+1,-2])
            
            #cumulative loglogistic distribution
            F = self.mortality_loglogistic(D_wm, params['m'], params['beta']) 
            
            #survival at time t/times[i]
            surv = y[0, -1] * (1.0 - F) * np.exp(-params['hb']*t)
            #replacement of the placeholder with the actual survival value
            y[i, -1] = surv
        return y
    
    @staticmethod
    def _rhs_jax(t, y, x_in, kd):
        D, = y
        C = x_in.evaluate(t)

        dD_dt = kd * (C - D)

        return (dD_dt, )

    @staticmethod
    def _solver_post_processing(results, t, interpolation, w, m, beta, hb, eps):
        # sum weighted damage
        D_sum = jnp.sum(results["D"] * w, axis=1)
        
        # calculate IT
        d_max = jnp.squeeze(jnp.array([jnp.max(D_sum[:i+1])+eps for i in range(len(t))]))
        F = jnp.where(d_max > 0, 1.0 / (1.0 + (d_max / m) ** -beta), 0)
        S = 1.0 * (jnp.array([1.0], dtype=float) - F) * jnp.exp(-hb * t)
        results["H"] = - jnp.log(S)
        results["survival"] = S
        results["exposure"] = jax.vmap(interpolation.evaluate)(t)

        return results

               
class RED_SD_IA(Mixture):
    '''
    the GUTS-RED-SD-IA model. This is the SD option of the independant action guts mixtures model.
    '''
    
    def __init__(self, num_expos = 1):
        super().__init__()     
        self.num_expos = num_expos
        self._numCLike = num_expos
        self._numSLike = num_expos + 2

        self._likelihood_func_jax = conditional_survival_hazard_error_model

        self.state_variables = {
            "D": {"dimensions": ["id", "time", "substance"], "observed": False, "y0": [0.0] * num_expos},
            "H_i": {"dimensions": ["id", "time", "substance"], "observed": False, "y0": [0.0] * num_expos},
            "H": {"dimensions": ["id", "time"], "observed": False},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        ###substance specific params
        for i in range(self.num_expos):
            #dominant rate constants
            self.params_info[f'kd{i+1}'] = {'name':f'kd{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance",}
            #killing rates
            self.params_info[f'b{i+1}'] = {'name':f'b{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance", }
            #medians of threshold distribution
            self.params_info[f'm{i+1}'] = {'name':f'm{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance", }
        #background mortality rate
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True,}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'damage {i+1}')
        for i in range(self.num_expos):
            self.comp_info.append(f'survival {i+1}')
        self.comp_info.append('background survival')
        self.comp_info.append('total survival')
    
    def _rhs(self, t, y, params, xc):

        dy = y.copy()
        
        #damages
        for i in range(self.num_expos):
            dy[i] = params[f'kd{i+1}'] * (xc[i](t) - y[i])
        
        hzd_rates = []
        #Hazard rates
        for i in range(self.num_expos):
            hzd_temp = params[f'b{i+1}'] * max(0, y[i] - params[f'm{i+1}'])
            hzd_rates.append(hzd_temp)

        #substance induced mortality
        for i in range(self.num_expos):
            dy[self.num_expos + i] = -hzd_rates[i] * y[self.num_expos + i]
                    
        #background mortality
        dy[self.num_expos + self.num_expos] = -params['hb'] * y[self.num_expos + self.num_expos]
        
        #placeholder for mulitplied survival
        dy[-1] = -99

        return dy   
    
    @staticmethod
    def _rhs_jax(t, y, x_in, kd, b, m):
        """Calculates the hazard of each component in x_in separately.
        """
        D, H = y
        dD_dt = kd * (x_in.evaluate(t) - D)

        # Hazard rates chemicals
        dH_i_dt = b * jnp.maximum(0.0, D - m)

        return dD_dt, dH_i_dt

    @staticmethod
    def _solver_post_processing(results, time, interpolation, hb):
        """
        Derivation:

        $$
        S = \prod_i S_i  
        log(S) = \sum_i \log(S_i)
        S = \exp(-H)
        \log(S) = sum_i \log(\exp(-H_i)) = sum_i -H_i
        S = exp(\sum_i -H_i) = \exp(- \sum_i H_i)
        $$

        cumulative hazard from the background hazard $dH_bg/dt = h_b$  is added as the
        closed form solution  $H_bg = h_b * t$
        """
        results["H"] = jnp.sum(results["H_i"], axis=1) + hb * time
        results["survival"] = jnp.exp(-results["H"])
        results["exposure"] = jax.vmap(interpolation.evaluate)(time)

        return results


    def _solve(self, params, y0, times, xc):
        
        #defining integrator, initial values and parameters
        int_ode = scipyInt.ode(self._rhs).set_integrator('lsoda')
        int_ode.set_initial_value(y0)
        int_ode.set_f_params(params, xc)
            
        #creating result data structures
        sols = [y0]
        
        #integration + insertion of the total survival into the result-list
        for t in times[1:]:
            int_ode.integrate(t)
            res = int_ode.y
            
            #multiplikation of survivals
            surv_ges = 1
            for i in range(self.num_expos+1):
                surv_ges = surv_ges * (res[self.num_expos + i] / sols[0][self.num_expos + i])
        
            res[-1] = surv_ges * sols[0][self.num_expos + self.num_expos]
            
            sols.append(res)
        #returns sols; contains [damage1,..., damageN, survival1, ..., survivalN, survivalHB, survivalGES] as columns
        return np.array(sols)                    
                    

class RED_IT_IA(Mixture):
    '''
    the GUTS-RED-IT-IA model. This is the IT option of the independant action guts mixtures model.
    '''    
    
    def __init__(self, num_expos = 1):
        super().__init__()     
        self._numSLike = num_expos + 2
        self._numCLike = num_expos
        self.num_expos = num_expos

        self._likelihood_func_jax = conditional_survival_hazard_error_model

        self.state_variables = {
            "D": {"dimensions": ["id", "time", "substance"], "observed": False, "y0": [0.0] * num_expos},
            "H": {"dimensions": ["id", "time",], "observed": False, },
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        #substance specific params
        for i in range(self.num_expos):
            #dominant rate constants
            self.params_info[f'kd{i+1}'] = {'name':f'kd{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance",}
            #medians of threshold distribution
            self.params_info[f'm{i+1}'] = {'name':f'm{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance",}
            #shape parameters for the distributions of thresholds
            self.params_info[f'beta{i+1}'] = {'name':f'beta{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": "substance",}
        #background mortality rate
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'damage {i+1}')
        for i in range(self.num_expos):
            self.comp_info.append(f'survival {i+1}')
        self.comp_info.append('background survival')
        self.comp_info.append('total survival')
            
    def _rhs(self, t, y, params, xc):

        dy = y.copy()
        
        #damages
        for i in range(self.num_expos):
            dy[i] = params[f'kd{i+1}'] * (xc[i](t) - y[i])
        
        return dy 
        
    def _solve(self, params, y0, times, xc):
         
        #solve uptake equation
        y = super()._solve(params, y0,  times, xc)
        
        for i, ystep in enumerate(y):
            if i == 0:
                continue
            
            t = times[i]
            
            Ds_max = []
            #maximum damages to this point
            for j in range(self.num_expos):
                D_max_temp = np.max(y[:i+1, j])
                Ds_max.append(D_max_temp)
                
            Fs = []
            #probability distributions
            for j in range(self.num_expos):
                F_temp = self.mortality_loglogistic(Ds_max[j], params[f'm{j+1}'], params[f'beta{j+1}'])
                Fs.append(F_temp)
            
            #background mortality
            surv_bg = y[0,self.num_expos + self.num_expos] * np.exp(-params['hb']*t)
            y[i,self.num_expos + self.num_expos] = surv_bg
            
            #survivals
            for j in range(self.num_expos):
                y[i, self.num_expos + j] = y[0, self.num_expos + j] * (1.0 - Fs[j])
                
            #multiplikation of survivals
            surv_ges = 1
            for j in range(self.num_expos + 1):
                surv_ges = surv_ges * (y[i, self.num_expos + j] / y[0, self.num_expos + j])
            
            #if there are different numbers of organisms per substance/treatment it must be defined which number 
            #should be used for the overall survival, atm. it is the number from the control groups
            y[i, -1] = surv_ges * y[0, self.num_expos]

        return y  
    
    @staticmethod
    def _rhs_jax(t, y, x_in, kd):
        D, = y
        C = x_in.evaluate(t)

        dD_dt = kd * (C - D)

        return (dD_dt, )

    @staticmethod
    def _solver_post_processing(results, t, interpolation, m, beta, hb, eps):
        # calculate IT and take the maximum over the time dimension for each damage
        d_max = jnp.squeeze(jnp.array([jnp.max(results["D"][:i+1], axis=0)+eps for i in range(len(t))]))
        
        # This equation is equivalent to the inverse of S described in Bart et al. 2021 for IT models
        # S = 1 / (1+d_max/m) ** beta = 1 - F 
        # with F = 1 - 1 / (1+d_max / m) ** -beta
        # note the difference in the minus sign in the exponent (beta)
        F = jnp.where(d_max > 0, 1.0 / (1.0 + (d_max / m) ** -beta), 0)

        # take the product of the CDFs of the individual threshold functions then multiply
        # with the survival function of the background hazard
        S = 1.0 * jnp.prod(jnp.array([1.0], dtype=float) - F, axis=1) * jnp.exp(-hb * t)
        results["H"] = - jnp.log(S)
        results["survival"] = S
        results["exposure"] = jax.vmap(interpolation.evaluate)(t)
        
        return results
    
class BufferGUTS(Reduced):
    extra_dim = "exposure_path"

class BufferGUTS_SD(BufferGUTS):
    '''
    The BufferGUTS SD (Stochastic Death) model for event based, discretized exposures as experienced by above-ground terrestrial arthropods
    '''
    def __init__(self):
        super().__init__()
        self._numCLike = 2
        self.comp_info = [
            'Buffer',
            'Damage',
            'Survival']

        self.state_variables = {
            "B": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "D": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "H": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model


        self.params_info['kd'] = {'name':'kd', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        self.params_info['b'] = {'name':'b', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-10, 'max':1.0e3, 'initial':1e-3, 'vary':True}
        self.params_info['eta'] = {'name':'eta', 'min':0, 'max':1.0e6, 'initial':60*24, 'vary':False}

    def _rhs(self, t, y, params, xc):
        xc = xc[0]
        dy = y.copy()
        # buffer
        if xc(t) > y[0]:
            speed = params['eta']
        else:
            speed = params["kd"]
        dy[0] = speed*(xc(t) -y[0])

        # Damage
        dy[1] = params['kd'] * (y[0] - y[1])

        # Hazard rate
        h_substance = params['b'] * max(0, y[1] - params['m'])
        h = h_substance + params['hb']

        # Survival
        dy[2] = -h * y[2]

        return dy

    @staticmethod
    def _rhs_jax(t, y, x_in, eta, kd, b, m, hb):
        B, D, H = y

        C = x_in.evaluate(t)

        buffer_filling_rate = eta * (C - B)
        buffer_depletion_rate = kd * (C - B)

        dB_dt = jnp.where(C >= B, buffer_filling_rate, buffer_depletion_rate)

        dD_dt = kd * (B - D)

        dH_dt = (b * jnp.maximum((D - m),jnp.array([0.0], dtype=float)) + hb)
        
        return dB_dt, dD_dt, dH_dt
        
    @staticmethod
    def _solver_post_processing(results, time, interpolation):
        results["exposure"] = jax.vmap(interpolation.evaluate)(time)
        results["survival"] = jnp.exp(-results["H"])
        return results

class BufferGUTS_IT(BufferGUTS):
    '''
    The BufferGUTS IT (Individual Tolerance) model for event based, discretized exposures as experienced by above-ground terrestrial arthropods
    '''
    def __init__(self):
        super().__init__()
        self._numCLike = 2
        self.comp_info = [
            'Buffer',
            'Damage',
            'Survival']    


        self.state_variables = {
            "B": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "D": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "H": {"dimensions": ["id", "time"], "observed": False},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model

        self.params_info['kd'] = {'name':'kd', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        self.params_info['beta'] = {'name':'beta', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-10, 'max':1.0e3, 'initial':1e-3, 'vary':True}
        self.params_info['eta'] = {'name':'eta', 'min':0, 'max':1.0e6, 'initial':60*24, 'vary':False}

    def _rhs(self, t, y, params, xc):
        xc = xc[0]
        dy = y.copy()

        if xc(t) > y[0]:
            speed = params["eta"]
        else:
            speed = params["kd"]
        dy[0] = speed*(xc(t) -y[0])

        # Damage
        dy[1] = params['kd'] * (y[0] - y[1])

        dy[2] = -99

        return dy

    @staticmethod
    def _rhs_jax(t, y, x_in, eta, kd):
        B, D = y 

        C = x_in.evaluate(t)

        buffer_filling_rate = eta * (C - B)
        buffer_depletion_rate = kd * (C - B)

        dB_dt = jnp.where(C >= B, buffer_filling_rate, buffer_depletion_rate)

        dD_dt = kd * (B - D)

        return dB_dt, dD_dt


    def _solve(self, params, y0, times, xc):        
        #Solve uptake equation
        y = super()._solve(params, y0,  times, xc)

        #Calculate individual tolerance survival
        for i, ystep in enumerate(y):
            if i == 0:
                continue

            t = times[i]

            #Maximum damage to this point
            D_wm = np.max(y[:i+1,1])

            #Calculate survival from ratio of max damage to tolerance threshold.
            F = self.mortality_loglogistic(D_wm, params['m'], params['beta'])
            surv = y[0, -1] * (1.0 - F) * np.exp(-params['hb']*t)
            y[i, -1] = surv

        return y
    
    @staticmethod
    def _solver_post_processing(results, t, interpolation, m, beta, hb, eps):
        """
        TODO: Try alternative formulation. This is computationally simpler and numerically
        more stable:
        log S = log 1.0 + log (1.0 - F) + log exp -hb * t = 0 + log (1.0 - F) - hb * t
        """

        d_max = jnp.squeeze(jnp.array([jnp.max(results["D"][:i+1])+eps for i in range(len(t))]))
        F = jnp.where(d_max > 0, 1.0 / (1.0 + (d_max / m) ** -beta), 0)
        S = 1.0 * (jnp.array([1.0], dtype=float) - F) * jnp.exp(-hb * t)
        results["H"] = - jnp.log(S)
        results["survival"] = S
        return results


class BufferGUTS_SD_DA(BufferGUTS):
    '''
    The BufferGUTS SD (Stochastic Death) model for multiple exposure paths for event based, discretized exposures as experienced by above-ground terrestrial arthropods.
    The Damage Addition (DA) variant assumes individual uptake kinetics per exposure path.
    '''
    def __init__(self, num_expos = 1):
        super().__init__()  
        self.num_expos = num_expos
        self._numCLike = 2 * num_expos + 1
        self._numSLike = 1


        self.state_variables = {
            "B": {"dimensions": ["id", "time", self.extra_dim], "observed": False, "y0": [0.0] * self.num_expos},
            "D": {"dimensions": ["id", "time", self.extra_dim], "observed": False, "y0": [0.0] * self.num_expos},
            "H": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model


        ###substance specific params
        for i in range(self.num_expos):
            #dominant rate constants
            if i == 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'initial':1.0, 'vary':False, "dims": self.extra_dim}
            if i > 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": self.extra_dim}
            
            self.params_info[f'kd{i+1}'] = {'name':f'kd{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": self.extra_dim}
                
        ###shared params
        #killing rate
        self.params_info['b'] = {'name':'b', 'min':1.0e-6, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #median of the threshold distribution 
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #background mortality rate
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-10, 'max':1.0e3, 'initial':1e-3, 'vary':True}
        self.params_info['eta'] = {'name':'eta', 'min':0, 'max':1.0e6, 'initial':60*24, 'vary':False}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'Buffer {i+1}')
        for i in range(self.num_expos):
            self.comp_info.append(f'Damage {i+1}')
        self.comp_info.append('Damage Sum')
        self.comp_info.append('Survival')
    
    def _rhs(self, t, y, params, xc):
        
        dy = y.copy()
        
        damage_sum = 0
        for i in range(self.num_expos):
            # Buffer
            if xc[i](t) > y[i]:
                speed = params["eta"]
            else:
                speed = params[f'kd{i+1}']
            dy[i] = speed*(xc[i](t) -y[i])

            # Damage
            new_damage = params[f'kd{i+1}'] * (y[i] - y[self.num_expos+i])
            dy[self.num_expos+i] = new_damage
            if i == 0:
                damage_sum += new_damage
            else:
                damage_sum += new_damage * params[f'w{i+1}']

        # Damagesum
        dy[-2] = damage_sum #np.sum(dy[self.num_expos:self.num_expos*2])
        
        #Hazard rate; dependant of total damage
        hc = params['b'] * max(0, y[-2] - params['m'])
        #Adding background mortality
        h = hc + params['hb']
        #survival
        dy[-1] = -h * y[-1]

        return dy

    @staticmethod
    def _rhs_jax(t, y, x_in, eta, kd, b, m, hb, w):

        B, D, H = y

        C = x_in.evaluate(t)

        buffer_filling_rate = eta * (C - B)
        buffer_depletion_rate = kd * (C - B)

        dB_dt = jnp.where(C >= B, buffer_filling_rate, buffer_depletion_rate)

        dD_dt = kd * (B - D)

        dH_dt = (b * jnp.maximum((jnp.sum(D * w) - m),jnp.array([0.0], dtype=float)) + hb)

        return dB_dt, dD_dt, dH_dt

    @staticmethod
    def _solver_post_processing(results, time, interpolation):
        results["exposure"] = jax.vmap(interpolation.evaluate)(time)
        results["survival"] = jnp.exp(-results["H"])
        return results

    
class BufferGUTS_SD_CA(BufferGUTS):
    '''
    The BufferGUTS SD (Stochastic Death) model for multiple exposure paths for event based, discretized exposures as experienced by above-ground terrestrial arthropods.
    The Concentration Addition (DA) variant assumes the same uptake kinetic for all exposure paths.
    '''
    def __init__(self, num_expos = 1):
        super().__init__()  
        self.num_expos = num_expos
        self._numCLike = num_expos + 2
        self._numSLike = 1

        self.state_variables = {
            "B": {"dimensions": ["id", "time", self.extra_dim], "observed": False, "y0": [0.0] * self.num_expos},
            "D": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "H": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model


        ###substance specific params
        for i in range(self.num_expos):
            if i == 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'initial':1.0, 'vary':False, "dims": self.extra_dim}
            if i > 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": self.extra_dim}
        ###shared params
        #dominant rate constants
        self.params_info['kd'] = {'name':'kd', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #killing rate
        self.params_info['b'] = {'name':'b', 'min':1.0e-6, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #median of the threshold distribution 
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #background mortality rate
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-10, 'max':1.0e3, 'initial':1e-3, 'vary':True}
        self.params_info['eta'] = {'name':'eta', 'min':0, 'max':1.0e6, 'initial':60*24, 'vary':False}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'Buffer {i+1}')
        self.comp_info.append('Buffer Sum')
        self.comp_info.append('Damage')
        self.comp_info.append('Survival')
    
    def _rhs(self, t, y, params, xc):
        
        dy = y.copy()
        
        buffer_sum = 0
        for i in range(self.num_expos):
            # Buffer
            if xc[i](t) > y[i]:
                speed = params["eta"]
            else:
                speed = params['kd']
            new_buffer = speed*(xc[i](t) -y[i])
            dy[i] = new_buffer

            if i == 0:
                buffer_sum += new_buffer
            else:
                buffer_sum += new_buffer * params[f'w{i+1}']

        # Buffer sum
        dy[-3] = buffer_sum

        dy[-2] = params['kd'] * (y[-3] - y[-2])
        #Hazard rate; dependant of total damage
        hc = params['b'] * max(0, y[-2] - params['m'])
        #Adding background mortality
        h = hc + params['hb']
        #survival
        dy[-1] = -h * y[-1]

        return dy

    @staticmethod
    def _rhs_jax(t, y, x_in, eta, kd, b, m, hb, w):
        """RHS accordint to Leo's buffer model"""

        B, D, H = y

        C = x_in.evaluate(t)

        buffer_filling_rate = eta * (C - B)
        buffer_depletion_rate = kd * (C - B)

        dB_dt = jnp.where(C >= B, buffer_filling_rate, buffer_depletion_rate)

        dD_dt = kd * (jnp.sum(B * w) - D)

        dH_dt = (b * jnp.maximum((D - m),jnp.array([0.0], dtype=float)) + hb)
        
        return dB_dt, dD_dt, dH_dt

    @staticmethod
    def _solver_post_processing(results, time, interpolation):
        results["exposure"] = jax.vmap(interpolation.evaluate)(time)
        results["survival"] = jnp.exp(-results["H"])
        return results

    
class BufferGUTS_IT_DA(BufferGUTS):
    '''
    The BufferGUTS IT (Individual Tolerance) model for multiple exposure paths for event based, discretized exposures as experienced by above-ground terrestrial arthropods.
    The Damage Addition (DA) variant assumes individual uptake kinetics per exposure path.
    '''
    
    def __init__(self, num_expos = 1):
        super().__init__()  
        self.num_expos = num_expos
        self._numCLike = 2 * num_expos + 1
        self._numSLike = 1

        self.state_variables = {
            "B": {"dimensions": ["id", "time", self.extra_dim], "observed": False, "y0": [0.0] * self.num_expos},
            "D": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "H": {"dimensions": ["id", "time"], "observed": False},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model

        ### substance specific parameters
        for i in range(self.num_expos):
            #dominant rate constants
            if i == 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'initial':1.0, 'vary':False, "dims": self.extra_dim}
            if i > 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": self.extra_dim}

            self.params_info[f'kd{i+1}'] = {'name':f'kd{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": self.extra_dim}

        ### shared parameters
        #median of the threshold distribution 
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #shape parameter
        self.params_info['beta'] = {'name':'beta', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #background mortality
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-10, 'max':1.0e3, 'initial':1e-3, 'vary':True}
        self.params_info['eta'] = {'name':'eta', 'min':0, 'max':1.0e6, 'initial':60*24, 'vary':False}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'Buffer {i+1}')
        for i in range(self.num_expos):
            self.comp_info.append(f'Damage {i+1}')
        self.comp_info.append('Damage Sum')
        self.comp_info.append('Survival')
            
    def _rhs(self, t, y, params, xc):

        dy = y.copy()
        damage_sum = 0
        for i in range(self.num_expos):
            # Buffer
            if xc[i](t) > y[i]:
                speed = params["eta"]
            else:
                speed = params[f'kd{i+1}']
            dy[i] = speed*(xc[i](t) -y[i])
            # Damage
            new_damage = params[f'kd{i+1}'] * (y[i] - y[self.num_expos+i])
            dy[self.num_expos+i] = new_damage
            if i == 0:
                damage_sum += new_damage
            else:
                damage_sum += new_damage * params[f'w{i+1}']

        # Damagesum
        dy[-2] = damage_sum

        return dy

    @staticmethod
    def _rhs_jax(t, y, x_in, eta, kd, w):
        B, D = y 

        C = x_in.evaluate(t)

        buffer_filling_rate = eta * (C - B)
        buffer_depletion_rate = kd *(C - B)

        dB_dt = jnp.where(C >= B, buffer_filling_rate, buffer_depletion_rate)
        
        dD_dt = jnp.sum(kd * (B - D) * w)

        return dB_dt, dD_dt

    @staticmethod
    def _solver_post_processing(results, t, interpolation, m, beta, hb, eps):
        """
        TODO: Try alternative formulation. This is computationally simpler and numerically
        more stable:
        log S = log 1.0 + log (1.0 - F) + log exp -hb * t = 0 + log (1.0 - F) - hb * t
        """
        results["exposure"] = jax.vmap(interpolation.evaluate)(t)

        d_max = jnp.squeeze(jnp.array([jnp.max(results["D"][:i+1])+eps for i in range(len(t))]))
        F = jnp.where(d_max > 0, 1.0 / (1.0 + (d_max / m) ** -beta), 0)
        S = 1.0 * (jnp.array([1.0], dtype=float) - F) * jnp.exp(-hb * t)
        results["H"] = - jnp.log(S)
        results["survival"] = S
        return results
        

class BufferGUTS_IT_CA(BufferGUTS):
    '''
    The BufferGUTS IT (Individual Tolerance) model for multiple exposure paths for event based, discretized exposures as experienced by above-ground terrestrial arthropods.
    The Concentration Addition (DA) variant assumes the same uptake kinetic for all exposure paths.
    '''
    
    def __init__(self, num_expos = 1):
        super().__init__()  
        self.num_expos = num_expos
        self._numCLike = num_expos + 2
        self._numSLike = 1


        self.state_variables = {
            "B": {"dimensions": ["id", "time", self.extra_dim], "observed": False, "y0": [0.0] * self.num_expos},
            "D": {"dimensions": ["id", "time"], "observed": False, "y0": 0.0},
            "H": {"dimensions": ["id", "time"], "observed": False},
            "survival": {"dimensions": ["id", "time"], "observed": True},
        }

        self._likelihood_func_jax = conditional_survival_hazard_error_model


        ### substance specific parameters
        for i in range(self.num_expos):
            if i == 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'initial':1.0, 'vary':False, "dims": self.extra_dim}
            if i > 0:
                self.params_info[f'w{i+1}'] = {'name':f'w{i+1}', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True, "dims": self.extra_dim}

        ### shared parameters
        #dominant rate constants
        self.params_info['kd'] = {'name':'kd', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #median of the threshold distribution 
        self.params_info['m'] = {'name':'m', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #shape parameter
        self.params_info['beta'] = {'name':'beta', 'min':1.0e-3, 'max':1.0e3, 'initial':1.0, 'vary':True}
        #background mortality
        self.params_info['hb'] = {'name':'hb', 'min':1.0e-10, 'max':1.0e3, 'initial':1e-3, 'vary':True}
        self.params_info['eta'] = {'name':'eta', 'min':0, 'max':1.0e6, 'initial':60*24, 'vary':False}
        #information about the compartments
        self.comp_info = []
        for i in range(self.num_expos):
            self.comp_info.append(f'Buffer {i+1}')
        self.comp_info.append('Buffer Sum')
        self.comp_info.append('Damage')
        self.comp_info.append('Survival')
            
    def _rhs(self, t, y, params, xc):

        dy = y.copy()
        buffer_sum = 0
        for i in range(self.num_expos):
            # Buffer
            if xc[i](t) > y[i]:
                speed = params["eta"]
            else:
                speed = params['kd']
            new_buffer = speed*(xc[i](t) -y[i])
            dy[i] = new_buffer
            if i == 0:
                buffer_sum += new_buffer
            else:
                buffer_sum += new_buffer * params[f'w{i+1}']

        # Buffer sum
        dy[-3] = buffer_sum

        dy[-2] = params['kd'] * (y[-3] - y[-2])

        return dy
        
    @staticmethod
    def _rhs_jax(t, y, x_in, eta, kd, w):
        B, D = y 

        C = x_in.evaluate(t)

        buffer_filling_rate = eta * (C - B)
        buffer_depletion_rate = kd * (C - B)

        dB_dt = jnp.where(C >= B, buffer_filling_rate, buffer_depletion_rate)

        dD_dt = kd * (jnp.sum(B * w) - D)

        return dB_dt, dD_dt
    
    def _solve(self, params, y0, times, xc):

        #sols/y now contains the values for [damage1,...,damageN, damagesGES] and placeholders for 
        #the survival which are calculated in the following since the dependance of the 
        #maximum total damage cannot be realised within the integrator
        y = super()._solve(params, y0,  times, xc)
        #determine the maximum total damage for every timestep and calculate the resulting survival
        for i, ystep in enumerate(y):
            if i == 0:
                continue
            
            t = times[i]

            #Maximum total damage to this point
            D_wm = np.max(y[:i+1,-2])
            
            #cumulative loglogistic distribution
            F = self.mortality_loglogistic(D_wm, params['m'], params['beta']) 
            
            #survival at time t/times[i]
            surv = y[0, -1] * (1.0 - F) * np.exp(-params['hb']*t)
            #replacement of the placeholder with the actual survival value
            y[i, -1] = surv

        return y

    @staticmethod
    def _solver_post_processing(results, t, interpolation, m, beta, hb, eps):
        """
        TODO: Try alternative formulation. This is computationally simpler and numerically
        more stable:
        log S = log 1.0 + log (1.0 - F) + log exp -hb * t = 0 + log (1.0 - F) - hb * t
        """
        results["exposure"] = jax.vmap(interpolation.evaluate)(t)

        d_max = jnp.squeeze(jnp.array([jnp.max(results["D"][:i+1])+eps for i in range(len(t))]))
        F = jnp.where(d_max > 0, 1.0 / (1.0 + (d_max / m) ** -beta), 0)
        S = 1.0 * (jnp.array([1.0], dtype=float) - F) * jnp.exp(-hb * t)
        results["H"] = - jnp.log(S)
        results["survival"] = S
        return results



from . import fit
from . import input_data
from . import model
from . import survival_models

__version__ = "1.6.0"


read_exposure_survival = input_data.read_exposure_survival

__all__ = [
    "read_exposure_survival",
]



import jax
import jax.numpy as jnp
import numpyro


maximum = jnp.frompyfunc(jnp.maximum, nin=2, nout=1, identity=None) # type: ignore


@jax.jit
def ffill_na(x, mask):
    """Forward-fill nan values in x. If a mask is provided, assume 

    Parameters
    ----------
    x : _type_
        _description_
    mask : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    if mask is None:
        mask = jnp.logical_not(jnp.isnan(x))
    idx = jnp.where(mask,jnp.arange(mask.shape[1]),0)
    idx_ = maximum.accumulate(jnp.array(idx), axis=1).astype(int)
    return x[jnp.arange(idx.shape[0])[:,None], idx_]


def conditional_survival_from_hazard(x, mask):
    """Calculates the conditional survival from cumulative hazard values. 
    This equation is used when survival is repeatedly observed over time and

    Parameters
    ----------
    x : np.ndarray[I,T, float]
        A 2-dimensional I x T array of cumulative hazards defined as H = -ln(S).
        I is the batch dimension and T is the time dimension

    mask : np.ndarray[I,T, bool]
        A 2-dimensional array of the same shape as x, taking True if the survival
        was observed for the given index (i,t) and taking False if survival was
        not observed for the given index (i,t).
    
    Returns
    -------
    out : np.ndarray[I,T, float]
        A matrix with conditional probabilities and nans in place where the 
        mask has nans. Output has the same shape as input.

    Example
    -------
    Calculation example from survival probabilities to conditional survival
    probabilities given some masked values.
    
    >>> S_i = np.array([
    >>>     [1.  , 0.75, 0.1 , 0.05, 0.01, 0.  ],
    >>>     [1.  , 0.75, 0.1 , 0.05, 0.01, 0.  ],
    >>>     [1.  , 0.75, 0.1 , 0.05, 0.01, 0.  ],
    >>>     [1.  , 0.75, 0.1 , 0.05, 0.01, 0.  ],
    >>>     [1.  , 0.75, 0.1 , 0.05, 0.01, 0.  ],
    >>> ])

    >>> mask_obs = np.array([
    >>>     [ True,  True,  True,  True,  True,  True],
    >>>     [ True, False,  True,  True,  True,  True],
    >>>     [ True, False, False,  True,  True,  True],
    >>>     [ True,  True, False,  True,  True,  True],
    >>>     [ True, False,  True, False,  True,  True],
    >>> ])

    >>> conditional_survival_from_hazard(-jnp.log(S_i), mask_obs)
    array([
       [1.0  0.75  0.133  0.5     0.2  0.0]
       [1.0   nan  0.1    0.5     0.2  0.0]
       [1.0   nan  nan    0.05    0.2  0.0]
       [1.0  0.75  nan    0.0667  0.2  0.0]
       [1.0   nan  0.1    nan     0.1  0.0]
    ])
    """

    # Append zeros (hazard) to the beginning of the array (this aligns with the
    # safe assumption that before the zeroth observation S(t=-1) = 1.0)
    x_ = jnp.column_stack([jnp.zeros_like(x[:, 0]), x])

    # also mask needs to be expanded accordingly
    mask_ = jnp.column_stack([jnp.ones_like(mask[:, 0]), mask])

    # fill NaNs with forward
    x_filled = ffill_na(x_, mask_)

    # calculate the conditional survival.
    conditional_survival = jnp.exp(x_filled[:, :-1] - (x_[:, 1:]))

    # add nans and return
    return jnp.where(
        mask, conditional_survival, jnp.nan
    )


def conditional_survival_hazard_error_model(theta, simulation_results, observations, indices, masks, make_predictions):
    """Computes the likelihood of observing K survivors at time t given that
    N survivors were alive at the previous observation t-1

    This is achieved by using the Cumulative hazard directly in the calculation
    of the conditional survival probability. Note that this equation can easily
    be adapted to compute the conditional lethality probability.

    $$\\Pr(t < T~|~t_0 < T) = \\frac{e^{-H(t)}}{e^{-H(t_0)}} = e^{-H(t) - (- H(t_0))} = e^{-H(t) + H(t_0)} = e^{-(H(t) - H(t_0))} = e^{H(t_0) - H(t)}$$

    Example
    -------

                          t0    t1   t2     t3  tinf 
    observations          10     8    5      2     0
    Pr(T > t)            1.0   0.8  0.5    0.2   0.0
    Pr(T > t | T > t-1) (1.0)  0.8  0.625  0.4   0.0


    If the experiment ends before the last organism has died, these information
    have to be included in he computation of the likelihood. For multinomial
    distributions this is done by computing the remaining (ubobserved) lethality
    until the end of the experiment, whihc is just the inverse of the number of
    survivors at the end of the experiment. Therefore, for conditional survival 
    the information about the number of alive organisms
    at the end of the experiment is contained in the last observation.      

    We compute an additional interval until the start of the experiment in order
    to align the results with the observations (which include t=0). Otherwise
    we would only compute T-1 intervals if we have T observations (including t=0).
    """

    H = simulation_results["H"]
    n_surv = observations["survivors_before_t"]
    S_mask = masks["survival"]
    obs_survival = observations["survival"]
    S_conditional = conditional_survival_from_hazard(H, S_mask)

    if make_predictions:
        obs_survival = None

    ID, T = S_mask.shape

    with numpyro.plate("time", size=T):
        with numpyro.plate("id", size=ID):
            numpyro.sample(
                "survival_obs", numpyro.distributions.Binomial(
                    probs=S_conditional, 
                    total_count=n_surv
                ).mask(S_mask), 
                obs=obs_survival
            )


def conditional_survival_error_model(theta, simulation_results, observations, indices, masks, make_predictions):
    """Note that the conditional survival error model will not work to generate
    posterior predictions for observational uncertainty out of the box
    
    This model can be used for SD and IT models, but for SD models, the 
    conditional_survival_hazard_model is preferrable.
    """
    # error model
    EPS = observations["eps"]
    S = jnp.clip(simulation_results["survival"], EPS, 1 - EPS) 

    S_mask = masks["survival"]
    n_surv = observations["survivors_before_t"]
    obs_survival = observations["survival"]
    
    S_ = ffill_na(S, S_mask)
    S_cond_na = S_[:, 1:] / S_[:, :-1]
    S_cond_na_ = jnp.column_stack([jnp.ones_like(S[:,[0]]), S_cond_na])
    S_cond_na__ = jnp.clip(S_cond_na_, EPS, 1 - EPS) 

    if make_predictions:
        obs_survival = None

    ID, T = S_mask.shape

    with numpyro.plate("time", size=T):
        with numpyro.plate("id", size=ID):
            numpyro.sample(
                "survival_obs", numpyro.distributions.Binomial(
                    probs=S_cond_na__, 
                    total_count=n_surv
                ).mask(S_mask), 
                obs=obs_survival
            )


def conditional_lethality_error_model(theta, simulation_results, observations, indices, masks, make_predictions):
    raise NotImplementedError(
        "Conditional lethality error model needs to forward fill survival " +
        "probabilities for missing (observation) values"
        ""
    )

    # error model
    EPS = observations["eps"]
    S = jnp.clip(simulation_results["S"], EPS, 1 - EPS) 

    # TODO work on this
    # Conditional survival may be wrong, but also maybe the data generation
    # fct may be wrong
    # a test showed that it looks preddy darn good. I guess this is because
    # it is the conditional probability to survive. Not the conditional
    # probability to die.
    S_cond = (S[:, :-1] - S[:, 1:]) / S[:, :-1]
    S_cond_ = jnp.column_stack([jnp.zeros_like(S[:,[0]]), S_cond])
    S_cond__ = jnp.clip(S_cond_, EPS, 1 - EPS) 
    
    n_surv = observations["survivors_before_t"]
    L_mask = masks["L"]
    L = observations["L"]

    if make_predictions:
        L = None

    numpyro.sample(
        "L_obs", numpyro.distributions.Binomial(
            probs=S_cond__, 
            total_count=n_surv
        ).mask(L_mask), 
        obs=L
    )


def multinomial_error_model(theta, simulation_results, observations, indices, masks, make_predictions):
    raise NotImplementedError(
        "Multinomial error model needs to forward fill survival " +
        "probabilities for missing (observation) values"
        ""
    )
    # error model
    EPS = observations["eps"]
    # TODO This should not be done before calculating the difference
    S = jnp.clip(simulation_results["S"], EPS, 1 - EPS) 

    # TODO work on this
    # This is not working with replicates
    s_probs = S[:, :-1] - S[:, 1:]
    s_probs_ = jnp.column_stack([jnp.zeros_like(S[:,[0]]), s_probs])
    
    L_mask = masks["L"]
    L = observations["L"]
    N = observations["n_subjects"]
    # N_ = jnp.broadcast_to(jnp.expand_dims(N, 1), L.shape)

    if make_predictions:
        L = None

    ID, T = L_mask.shape

    with numpyro.plate("id", ID):
        numpyro.sample(
            "L_obs", numpyro.distributions.Multinomial(
                probs=s_probs_, 
                total_count=N
            ).mask(L_mask), 
            obs=L
        )


'''
This file contains the datastructure mempy.fit, which can be used to fit models to input-data
and which must be passed as an attribute to functions performing simulation or validation.

Classes
----------
Fit
    __init__()
        Initialises a mempy.fit-object.
    _check_overwrite()
        Warns the user before they overwrite the information of an existing fit by performing a new fit.
    pretty_print()
        Prints all available information on the fit performed with this instance in a comprehensive way to the console.
    serialize()
        Serializes this Fit instance using the pickle library to ensure future availability of the achieved results.
    visualize()
        Plots the model solutions with the fitted parameter values and the input-survival-data
    non_lmfitize()
        Converts a lmfit.Parameters-object in to a dictionary of parameter values usable by the mempy.model classes and passes it to their negloglike function.
    nelder_mead_simplex()
        The Nelder-Mead-simplex fitting method powered by the lmfit-library.
    mcmc()
        Fitting via lmfit + emcee solver. Can either be used to only calucalte the uncertainties as the 2.5 and 97.5 percentiles of the flatchain or
        for parameter values as well (use with caution and see the emcee python package for details)
    simple_parameter_space_explorer()
        Fits a model to data using a simple random parameter explorer and the nelder_mead_simplex()-function
    
Functions
----------
deserialize()
    recover a serialized fit-object using pickle
deserialize_many()
    recover all fits in a given directory using pickle
'''
import lmfit
import pandas as pd
from tabulate import tabulate
import mempy.input_data as input_data
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import pickle
import os
import emcee
import math
import corner
    
class Fit:

    def __init__(self):
        '''
        Initialising information about the performed fit.
        '''
        self.method = None
        self.params = None
        self.params_info = None
        self.residual = None
        self.uncertainties = None
        self.fit_info = None
        self.goodness = None
        self.hb_fix = None
        self.data = None
        self.data_info = None
        self.model = None
        self.flatchain = None
        self.uncertainties_method = None


    def _check_overwrite(self, overwrite):
        '''
        This method warns the user before they overwrite the information of an existing fit by performing a new fit.

        Parameters
        ----------
        overwrite: boolean
            Wheter or not the user allows overwriting the information in this fit.Fit instance.
        '''
        if self.method != None:
            if not overwrite:
                raise PermissionError("This fit-instance has already been used to perform a fit. Create a new mempy.fit.Fit-object or set overwrite to True.")


    def pretty_print(self):
        '''
        Prints all available information on the fit performed with this instance in a comprehensive way to the console.
        ''' 
        print(f'Model: {self.model.__class__.__name__}')
        print(f'Number of exposures: {self.model.num_expos}\n')

        print(' Best-Fit parameter values'.upper())
        keys = []
        values = []
        
        for key, value in self.params.items():
            keys.append(key)
            values.append(value)

        if self.uncertainties == None:
            list = [keys, values]
            print(tabulate(list, headers = 'firstrow', tablefmt="heavy_grid"))
            print('Uncertainties were not calculated.')
        else:
            upper = []
            lower = []
            #upper = []
            for key, value in self.params.items():
                if key in self.uncertainties:
                    lower.append(self.uncertainties[key]['lower'])
                    upper.append(self.uncertainties[key]['upper'])
                else:
                    lower.append['nAn']
                    upper.append['nAn']
            list = [keys, values, lower, upper]
            print(tabulate(list, headers = 'firstrow', tablefmt="heavy_grid", showindex=['Value', 'lower', 'upper']))
        
        print('\n', 'Fit-algorithm  Info'.upper())

        print(f'Method: {self.method}')
        hbf = {True: 'was', False: 'was not'}
        print(f'The background mortality {hbf[self.hb_fix]} priorly fitted to control-data.')

        keys = []
        values = []
        for key, value in self.fit_info.items():
            keys.append(key)
            values.append(value)
        list = [keys, values]
        print(tabulate(list, headers = 'firstrow', tablefmt="heavy_grid"))
        print(f'Value of the target function: {self.residual}')

        print('\n', 'Parameter ranges')
        keys = []
        min = []
        max = []
        initial = []
        for key, value in self.model.params_info.items():
            keys.append(key)
            min.append(value['min'])
            max.append(value['max'])
            initial.append(value['initial'])
        list = [keys, min, max, initial]
        print(tabulate(list, headers = 'firstrow', tablefmt='heavy_grid', showindex=['min','max','initial']))

        print('\n', 'Goodness of fit'.upper())
        keys = []
        values = []
        for key, value in self.goodness.items():
            keys.append(key)
            values.append(value)
        list = [keys, values]
        print(tabulate(list, headers = 'firstrow', tablefmt="heavy_grid"))

        print('\n', 'Data info'.upper())
        keys = []
        info = []
        for key, inf in self.data_info.items():
            keys.append(key)
            info.append(inf)

        list = [keys, info]
        print(tabulate(list, headers = 'firstrow', tablefmt="heavy_grid"))

    def serialize(self, dir, filename):
        '''
        Serializes this Fit instance using the pickle library to ensure future availability of the achieved results.

        Parameters
        ----------
        dir: string
            The directory in which the instance should be saved. Is created if it does not exist.
        filename: string
            The name of the .pickle-file in which this instance should be saved.
        '''
        if not os.path.exists(dir):
            os.makedirs(dir)
        
        with open(dir + filename + '.pickle', "wb") as file:
            pickle.dump(self, file)

    

    def visualize(self, share_axis = True, suptitle = 'Fit to training data'):
        '''
        Plots the model solutions with the fitted parameter values and the input-survival-data

        Parameters
        ----------
        share_axis: boolean
            Wheter or not the plots for each compartment should share their y-axis to achieve better visual comparability.
        suptitle: string
            The Title for the figure.
        '''
        self.model.plot(self.params, self.data['exposure'], observed_data = self.data['survival'], share_axis = share_axis, suptitle = suptitle)

    
    def non_lmfitize(self, params, datasets, model):
        '''
        Converts a lmfit.Parameters-object in to a dictionary of parameter values usable by the mempy.model classes and passes it to their negloglike function.

        This is to ensure, that the mempy.model classes can be entirely independant any specific library (lmfit) to allow the implementation
        of different fitting-methods that are not supported in lmfit.

        Parameters
        ----------
        params: lmfit.Parameters
            A lmfit.Parameters object to be evaluated with the model.
        datasets: dict
            The training data to compare the model results to.
        model: mempy.model.Model
            The bottom level mempy.model instance implementing the model ODEs

        Returns
        ----------
        result: float
            The resulting negative logarithmic likelihood as calculated by the negloglike-method of the passed model instance.
        '''
        params_dict = params.valuesdict()
        result = model.negloglike(params_dict, datasets)
        return result


    def nelder_mead_simplex(self, model, exposure_funcs, survival_data, hb_fix = False, maxiter = None, printres = True, visualize = True, overwrite = False):
        '''
        The Nelder-Mead-simplex fitting method powered by the lmfit-library.

        Parameters
        ----------
        model: mempy.model.Model
            The bottom level model instance to be fitted
        exposure_funcs: dict
            Dictionary of functions representing the time dependant external exposure
        survival_data: pandas.DataFrame
            The survival_data to which the model should be fitted
        hb_fix: boolean
            Wheter or not the background mortality should priorly be fitted solely to the control datasets and subsequently fixed to that value.
        printres: boolean
            Wheter or not the results of this fit should be printed to the console
        visualize: boolean
            Wheter or not the model should be plotted with the resulting parameter values
        overwrite: boolean
            Wheter or not the results saved in an instance can be overwritten by performing a second fit.
        '''
        self._check_overwrite(overwrite)

        datasets = input_data.create_datasets(exposure_funcs, survival_data)

        lmfit_params = lmfit.Parameters()
        for p in model.params_info:
            lmfit_params.add(model.params_info[p]['name'], 
                            min = model.params_info[p]['min'], 
                            max = model.params_info[p]['max'], 
                            value = model.params_info[p]['initial'], 
                            vary = model.params_info[p]['vary'])
            
        if hb_fix:

            control_datasets, datasets = input_data.split_control_treatment_dataset(datasets)

            for p in lmfit_params:
                lmfit_params[p].vary = False
            lmfit_params['hb'].vary = True

            minimizer = lmfit.Minimizer(self.non_lmfitize, lmfit_params, fcn_args = (control_datasets, model))
            result = minimizer.minimize(method='nelder', params=lmfit_params)

            lmfit_params = result.params

            for p in lmfit_params:
                lmfit_params[p].vary = True
            lmfit_params['hb'].vary = False
 
        minimizer = lmfit.Minimizer(self.non_lmfitize, lmfit_params, fcn_args = (datasets, model))
        result = minimizer.minimize(method='nelder', params=lmfit_params, options={'maxiter':maxiter})

        # Saving the results
        self.method = "Nelder-Mead-Simplex" 
        self.params = result.params.valuesdict() 
        self.params_info = model.params_info
        self.residual = result.residual
        self.uncertainties = None
        self.fit_info = {
            'success'   : result.success,
            'n_f_ev'    : result.nfev,
            'n_vary'    : result.nvarys
            }
        self.goodness = {
            'aic'       : result.aic,
            'bic'       : result.bic,
            'chi_sq'    : result.chisqr,
            'nrmse'     : model.nrmse(self.params, exposure_funcs, survival_data, with_akaike = False)
            }
        self.hb_fix = hb_fix
        self.model = model
        self.data = {
            'exposure': exposure_funcs,
            'survival': survival_data
        }
        self.data_info = {
            'num_expos': model.num_expos,
            'num_treatments': len(survival_data),
            'starttime' : survival_data.index[0],
            'endtime' : survival_data.index[-1],
            'num_individuals' : survival_data[survival_data.keys()[0]][0]
            }

        if printres:
            self.pretty_print()
        if visualize:
            self.visualize()

    
    def mcmc(self, model, exposure_funcs, survival_data, hb_fix=True, nsteps=10000, thin=1, nwalkers=None, uncertainties_only = True, progressbar=True, printres = True, visualize = True, overwrite = False):
        '''
        Fitting via lmfit + emcee solver. Can either be used to only calucalte the uncertainties as the 2.5 and 97.5 percentiles of the flatchain or
        for parameter values as well (see the emcee python package)

        Parameters
        ----------
        model: mempy.model.Model
            The bottom level model instance to be fitted
        exposure_funcs: dict
            Dictionary of functions representing the time dependant external exposure
        survival_data: pandas.DataFrame
            The survival_data to which the model should be fitted
        hb_fix: boolean
            Wheter or not the background mortality should priorly be fitted solely to the control datasets and subsequently fixed to that value.
        nsteps: int
            The amount of steps for each walker
        thin: int
            only accepts one in every *thin* samples (see the emcee python package)
        nwalkers: int
            Should be set so nwalkers >> nvarys, where nvarys are the number of parameters being varied during the fit. 
            'Walkers are the members of the ensemble. They are almost like separate Metropolis-Hastings chains but, of course, the proposal 
            distribution for a given walker depends on the positions of all the other walkers in the ensemble'. - from the emcee webpage.
        uncertainties_only: boolean
            Wheter or not only uncertainties should be calculated for already fitted parameters
        progressbar: boolean
            Wheter or not a progressbar should be shown
        printres: boolean
            Wheter or not the results of this fit should be printed to the console
        visualize: boolean
            Wheter or not the model should be plotted with the resulting parameter values
        overwrite: boolean
            Wheter or not the results saved in an instance can be overwritten by performing a second fit.        
        '''
        if not uncertainties_only:
            self._check_overwrite(overwrite)
        else:
            if self.method == None:
                raise PermissionError("Cannot calculate uncertainties only since best fit parameter values have not yet been calucalted.")


        datasets = input_data.create_datasets(exposure_funcs, survival_data)
        lmfit_params = lmfit.Parameters()

        if not uncertainties_only:

            for p in model.params_info:
                lmfit_params.add(model.params_info[p]['name'], 
                                min = model.params_info[p]['min'], 
                                max = model.params_info[p]['max'], 
                                value = model.params_info[p]['initial'], 
                                vary = model.params_info[p]['vary'])

            if hb_fix == True:
                control_datasets, datasets = input_data.split_control_treatment_dataset(datasets)

                for p in lmfit_params:
                    lmfit_params[p].vary = False
                lmfit_params['hb'].vary = True

                minimizer = lmfit.Minimizer(self.non_lmfitize, lmfit_params, fcn_args = (control_datasets, model))
                result = minimizer.minimize(method='nelder', params=lmfit_params)

                lmfit_params = result.params

                for p in lmfit_params:
                    lmfit_params[p].vary = True
                lmfit_params['hb'].vary = False
            
        else:
            lmfit_params = lmfit.Parameters()
            for p in model.params_info:
                lmfit_params.add(model.params_info[p]['name'], 
                                min = model.params_info[p]['min'], 
                                max = model.params_info[p]['max'], 
                                value = self.params[p], 
                                vary = model.params_info[p]['vary'])
                
        def objective(*args, **kwargs):
            return -model.negloglike(*args, **kwargs)

        if not nwalkers:
            nwalkers = 2*len([n for n, p in lmfit_params.items() if p.vary])

        print(f'Commencing MCMC with {nwalkers} walkers, {nsteps} steps')

        mini = lmfit.Minimizer(objective, lmfit_params, fcn_args=(datasets,))

        result = mini.emcee(burn=int(nsteps/2), steps=nsteps, thin=thin, nwalkers=nwalkers, 
                         ntemps=1, params=lmfit_params, progress=progressbar)
        
        unc = {}
        for p in lmfit_params:
            unc[p] = {'lower':1,'upper':1}
            unc[p]['lower'] = np.percentile(result.flatchain[p],2.5)
            unc[p]['upper'] = np.percentile(result.flatchain[p],97.5)

        if not uncertainties_only:
            # Saving the results
            self.method = "MCMC" 
            self.params = result.params.valuesdict() 
            self.params_info = model.params_info
            self.residual = result.residual
            self.uncertainties = unc
            self.flatchain = result.flatchain
            self.fit_info = {
                'success'   : result.success,
                'n_f_ev'    : result.nfev,
                'n_vary'    : result.nvarys
                }
            self.goodness = {
                'aic'       : result.aic,
                'bic'       : result.bic,
                'chi_sq'    : result.chisqr,
                'nrmse'     : model.nrmse(self.params, exposure_funcs, survival_data, with_akaike = False)
                }
            self.hb_fix = hb_fix
            self.model = model
            self.data = {
                'exposure': exposure_funcs,
                'survival': survival_data
            }
            self.data_info = {
                'num_expos': model.num_expos
                }
        else:
            self.uncertainties_method = 'mcmc'
            self.uncertainties = unc
            self.flatchain = result.flatchain

        def plot_mcmc(flatchain, res, quantiles=[0.025, 0.5, 0.975]):
            labels = flatchain.columns 
            c = corner.corner(flatchain, labels=labels,
                truths=[res[par] for par in res if par in flatchain.columns], title_fmt=".2E", 
                        verbose=True, show_titles=True, quantiles=quantiles)

        if printres:
            self.pretty_print()
        if visualize:
            self.visualize()
            plot_mcmc(self.flatchain, self.params)

    
    def simple_parameter_space_explorer(self, model, exposure_funcs, survival_data, hb_fix = False, combinations=100, pre_maxiter = 60, final_maxiter = None, printres = True, visualize = True, overwrite = False):
        '''
        Fits a model to data using a simple random parameter explorer and the nelder_mead_simplex()-function
        
        Parameters
        ----------
        model: mempy.model.Model
            The bottom level model instance to be fitted
        exposure_funcs: dict 
            dict containing the exposure functions (see mempy.input)
        survival_data: pandas.DataFrame
            DataFrame containing the treatments as columns and times as index (see mempy.input)
        hb_fix: boolean
            wheter or not the background mortality should priorly be fitted solely to the control datasets and subsequently fixed to that value.
        combinatinos: int
            the amount of initial parameter sets to be tested
        pre_maxiter: int
            the amount of steps in the minimization-process for the initial parameter sets
        final_maxiter: int
            the amount of steps in the minimization-process for the best of the initial parameter sets
        printres: boolean
            Wheter or not the results of this fit should be printed to the console
        visualize: boolean
            Wheter or not the model should be plotted with the resulting parameter values
        overwrite: boolean
            Wheter or not the results saved in an instance can be overwritten by performing a second fit.
        '''
        self._check_overwrite(overwrite)

        def random_combinations(params_info, combinations, hb_fix):
            #Logarithmically draws N combinations of values for the parameters
            param_combinations = []
            for _ in range(combinations):
                comb = {}
                for param in params_info:
                    if param == "hb" and hb_fix:
                        continue

                    log_max = np.log10(params_info[param]['max'])
                    log_min = np.log10(params_info[param]['min'])

                    rand = np.random.random()
                    rand_val = 10 ** (rand * (log_max - log_min) + log_min)
                    comb[param] = rand_val
                param_combinations.append(comb)
            return param_combinations 
        
        if combinations > 0:    
            param_combinations = random_combinations(model.params_info, combinations, hb_fix)
            param_exp_list = []
            for i, param_combi in enumerate(param_combinations):
                print(f'Starting parameter combination: {i}', end = '\r')
                for key in model.params_info:
                    if key in param_combi:
                        model.params_info[key]['initial'] = param_combi[key]

                self.nelder_mead_simplex(model, exposure_funcs, survival_data, hb_fix = hb_fix, maxiter = pre_maxiter, printres = False, visualize= False, overwrite = True)
                param_exp_list.append((self.residual, self.params))
            
            sort_params = sorted(param_exp_list, key=lambda tup: tup[0])
            
            for k in model.params_info:
                model.params_info[k]['initial'] = sort_params[0][1][k]
            
            print(f'\nStarting final minimization with optimal initial parameter set {sort_params[0][1]} with residual {sort_params[0][0]}.')
        self.nelder_mead_simplex(model, exposure_funcs, survival_data, hb_fix = hb_fix, maxiter = final_maxiter, printres = printres, visualize= visualize, overwrite = True)

        self.method = 'simple_parameter_space_explorer'



def deserialize(dir, filename):
    '''
    Recover a serialized fit-object using pickle

    Parameters
    ----------
    dir: string
        The directory containing the serialized fit-object to be recovered.
    filename: string
        The .pickle-file containg the fit-object to be recovered.

    Returns
    ----------
    ret: mempy.fit.Fit
        The deserialized fit-instance
    '''
    with open(dir + filename, "rb") as file:
        ret = pickle.load(file)
    return ret

def deserialize_many(dir):
    '''
    Recover all serialized fit-objects in a given directory using pickle

    Parameters
    ----------
    dir: string
        The directory containing the serialized fit-objects to be recovered.

    Returns
    ----------
    fits: list
        A list containing all the deserialized fit-objects
    '''
    fits = []
    for filename in os.listdir(dir):
        with open(dir+filename, "rb") as file:
            if ".pickle" in filename:
                fits.append(pickle.load(file))
    return fits



'''
This file contains functions for handling input data, i.e. reading from files and converting to 
datastructures usable by this implementation.

Classes
----------
None

Functions
----------
multi_interp_treatment_generator()
    Creates interpolating treatment functions from exposure data.
create_datasets()
    Converts input-data to a form structured around the individual treatments.
read_exposure_survival()
    Reads a xlsx file with the exposure and survival data and converts it into useable datastructures.
get_xc()
    Extracts every exposure-function for a given treatment.
split_control_treatment_dataset():
    Splits a dataset into control and treatment datasets.
plot()
    Visualizes the Exposure- and Survival-data as imported by the input_data.read_exposure_survival()-function.
'''
import os
import pandas as pd
from scipy import interpolate
import matplotlib.pyplot as plt
import numpy as np

def multi_interp_treatment_generator(treatments):
    '''
    Creates interpolating treatment functions from exposure data.
    
    Parameters
    ----------
    treatments: pandas.DataFrame
        DataFrame containing the exposure concentrations with the treatments as columns and the timepoints as rows.

    Returns
    ----------
    treatment_funcs: dict
        Dictionary with the treatment-names as keys and the corresponding functions as values. 
    '''
    treatment_funcs = {}
    for treat in treatments:
        t = treatments[treat].index
        conc = treatments[treat].values
        conc1 = conc[~np.isnan(conc)]
        t = t[~np.isnan(conc)]
        s = interpolate.InterpolatedUnivariateSpline(t, conc1, k=1)
        s.__name__ = f'treatment_{treat.lower()}'
        treatment_funcs[treat] = s
    return treatment_funcs


def create_datasets(exposure_funcs, survival_data, treatment_list=None):
    '''
    Converts input-data to a form structured around the individual treatments.

    Parameters
    ----------
    exposure_funcs: dictionary
        Dict of exposures and treatments containing functions representing the external concentrations.
    survival_data: pandas.Dataframe
        Pandas.DataFrame containing the survival for the treatments and timepoints.
    treatment_list: list, optional
        List of strings to name the treatments. If None, the headers of the survival_data-dataframe are used.

    Returns
    ----------
    datasets: list
        List of datasets containing tuples with all exposures and the corresponding survival per treatment.
        In the following Form:
        [
            (exposure_1_treatment_1, exposure_2_treatment_1, ... , exposure_n_treatment_1, survival_treatment_1),
            (exposure_1_treatment_2, exposure_2_treatment_2, ... , exposure_n_treatment_2, survival_treatment_2),
            ...
            (exposure_1_treatment_m, exposure_2_treatment_m, ... , exposure_n_treatment_m, survival_treatment_m)
        ]
    '''
    datasets = []
    if treatment_list == None:
        survival_columns = list(survival_data.keys())
    else:
        survival_columns = treatment_list
    for treatment in survival_columns:
        current_exposure = []
        for sub in exposure_funcs:
            current_exposure.append(exposure_funcs[sub][treatment])
        datasets.append((*current_exposure, survival_data[treatment].dropna()))
    return datasets
    

def read_exposure_survival(
    path, 
    file_name, 
    exposure_name="Exposure", 
    survival_name="Survival", 
    info_name="Info", 
    with_info_dict=False, 
    with_raw_exposure_data=False, 
    visualize=False
):
    
    """
    Reads a xlsx file with the exposure and survival data and converts it into useable datastructures.
    
    Parameters
    --------
    path : str
        Path to the file
    file_name : str
        File name of the xlsx file with the survival and exposure data
    exposure_name : str, default='Exposure'
        Name of the datasheet(s) containing the exposure data
    survival_name : str, default='Survival'
        Name of the datasheet containing the survival data
    info_name : str, default='Info'
        Name of the datasheet containing additional informations
    with_info_dict : bool, default=False
        If True the information datasheet is extracted and converted into a dictionary
    with_raw_exposure_data : bool, default=False
        If the raw exposure data should be returned
    visualize : bool, default=False
        If True the imported data is visualized in two plots
    
    Returns
    --------
    exposure_funcs : dict 
        Dictionairy containing the exposure functions
    survival_data : DataFrame
        Pandas DataFrame containing the treatments as columns and times as index
    num_expos : int
        Number of different expositions that the organisms are subject to. This will be one in the standard guts models but e.g. the number of substances
        used in the exposure when considering a guts-mixtures model.
    info_dict : dict, optional
        Contents of the info datasheet, only returned if 'with_info_dict'=True
    """
    file = pd.ExcelFile(os.path.join(path, file_name))
    exposure_sheets = [exp for exp in file.sheet_names if exposure_name in exp]
    num_expos = len(exposure_sheets)
    if num_expos < 1:
        raise ValueError(f'No sheet named {exposure_name} in the Excel file')
    else:
        exposure_funcs = {}
        raw_exposure_data = {}
        for exposure_sheet in exposure_sheets:
            exposure_data = file.parse(sheet_name=exposure_sheet,index_col=0)
            exposure_funcs[exposure_sheet] = multi_interp_treatment_generator(exposure_data)
            raw_exposure_data.update({exposure_sheet: exposure_data})
    survival_data = file.parse(sheet_name=survival_name,index_col=0)

    if visualize:
        plot(exposure_funcs, survival_data)
    
    return_data = [exposure_funcs, survival_data, num_expos]

    if with_info_dict:
        df_info = file.parse( header=None, sheet_name=info_name)
        info_array = df_info.to_numpy()
        info_dict = {info[0]:info[1:] for info in info_array}
        return_data += [info_dict]

    if with_raw_exposure_data:
        return_data += [raw_exposure_data]

    return tuple(return_data)
    

def get_xc(treatment, exposure_funcs):
    '''
    Extracts every exposure-function for a given treatment.

    Parameters
    ----------
    treatment: String
        Name of the treatment to be handled.
    exposure_funcs: dict
        Dictionary of the exposures and treamtents containing the exposure-functions.
    
    Returns
    ----------
    xc: Tuple
        Tuple containing all exposure-functions for the given treatment.
    '''
    xc_list = []
    for exposure in exposure_funcs:
        xc_list.append(exposure_funcs[exposure][treatment])
    xc = tuple(xc_list)
    return xc


def split_control_treatment_dataset(datasets):
    """
    Splits a dataset into control and treatment datasets

    Parameters
    ----------
    datasets: dict
        A dictionary of all datasets that should be categorized.

    Returns
    ----------
    control_datasets: dict
        A dictionary containing the control datasets
    treat_datasets: dict
        A dictionary containing all other datasets
    """
    control_datasets = []
    treat_datasets = []
    for dataset in datasets:
        if "control" in dataset[-1].name.lower().strip():
            control_datasets.append(dataset)
        else:
            treat_datasets.append(dataset)

    return control_datasets, treat_datasets 


def plot(exposure_funcs, survival_data):
    """
    Visualizes the Exposure- and Survival-data as imported by the input_data.read_exposure_survival()-function.

    Parameters
    ----------
    exposure_funcs: dict
        Dictionary of functions representing the time dependant external exposure to be visualized
    survival_data: pandas.DataFrame
        Table of survival-data to be visualized
    """
    fig, ax = plt.subplots()
    times = survival_data.index
    for col in survival_data.columns:
        ax.plot(times, survival_data[col], label = col)
    ax.legend()
    ax.set(xlabel='Time', ylabel='Survival', title='Observed survival per treatment over Time')
    plt.show()

    fig, ax = plt.subplots()
    for substance in exposure_funcs.keys():
        for treatment in exposure_funcs[substance].keys():
            times = np.linspace(exposure_funcs[substance][treatment].get_knots()[0],exposure_funcs[substance][treatment].get_knots()[-1], 1000)
            ax.plot(times, exposure_funcs[substance][treatment](times), label = treatment)
    ax.legend()
    ax.set(xlabel='Time', ylabel='Concentration', title='Exposure-Concentrations/Treatments over Time')
    plt.show()
